==> maxdb-source-7_6_00_09.tgz: V76_00_09/MaxDB_ORG/sys/src/SAPDB/FreeBlockManagement/FBM_ClusteredDataVolume.hpp <==
/*!

  @file           FBM_ClusteredDataVolume.hpp
  @author         Henrik
  @ingroup        FBM

\if EMIT_LICENCE
    ========== licence begin  GPL
    Copyright (c) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
\endif
*/

#ifndef FBM_CLUSTEREDDATAVOLUME_H
#define FBM_CLUSTEREDDATAVOLUME_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Container/Container_Vector.hpp"
#include "FreeBlockManagement/FBM_IDataVolume.hpp"
#include "FreeBlockManagement/FBM_BlockStateList.hpp"
#include "FreeBlockManagement/FBM_ClusteredFreeList.hpp"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
  @class          FBM_ClusteredDataVolume
  @brief          representation of one data volume
 */
class FBM_ClusteredDataVolume : public FBM_IDataVolume
{
public:
    typedef Container_Vector<SAPDB_UInt2>  FBM_CounterList;

private: 

    const IOMan_BlockNo           m_LastBlock;
    const RTE_VolumeAccessMode m_VolMode;

    IOMan_BlockCount      m_NumBlocksToAddUntilOptIsReached;
    IOMan_BlockNo         m_ActBlockNoForBackup;            
    IOMan_BlockCount      m_NumBlocksMarkedForBackup;
    IOMan_BlockCount      m_NumBlocksFreeAfterSVP;
    IOMan_BlockCount      m_NumBlocksUsed;
    IOMan_BlockNo         m_LastUsedBlock;
    IOMan_BlockNo         m_LastFragmentedBlock;
    FBM_BlockStateList    m_BlockStateList;         
    FBM_ClusteredFreeList m_ClusterFreeList;
    SAPDB_Bool            m_VolumeIsInCompressingMode;
    SAPDB_Bool            m_IsToDrop;

    IOMan_BlockNo GetOptLastUsedBlock() const
    {
        return SAPDB_MIN((IOMan_BlockNo)(m_NumBlocksUsed + (m_NumBlocksUsed >> 2)),(m_LastBlock));
    }

    void FindPairOfFreeBlocks (
        IOMan_BlockNo  StartSearchBlockNo, 
        SAPDB_Bool &bFreePairFound,
        IOMan_BlockNo &FreePairBlockNo,
        IOMan_BlockNo &StartBlockNoForNextSearch) const;

    void pSetBlockStateMarkedForCompressionToFree (IOMan_BlockNo BlockNo)
    {
        /* check if the FBM is in the compression mode where this state can only exist */ 
        SAPDBERR_ASSERT_STATE (m_VolumeIsInCompressingMode);
        
        /* check if the transition is correct */
        SAPDBERR_ASSERT_STATE ((GetBlockStateAfterBackup (BlockNo) == BlockState_Free) &&
            (GetBlockState            (BlockNo) == BlockState_BackUp));
        
        /* update counter of blocks marked for back up */
        --m_NumBlocksMarkedForBackup;
        
        /* set block state to free */
        m_BlockStateList.SetBlockStateAfterBackup (BlockNo, BlockState_Free);
        SetBlockStateToFree (BlockNo);
    }
    
    void pSetBlockStateMarkedForCompressionToOccupied  (IOMan_BlockNo BlockNo)
    {
        /* check if the FBM is in the compression mode where this state can only exist */ 
        SAPDBERR_ASSERT_STATE (m_VolumeIsInCompressingMode);
        
        /* check if the transition is correct */
        SAPDBERR_ASSERT_STATE ((GetBlockStateAfterBackup (BlockNo) == BlockState_BackUp) &&
            (GetBlockState            (BlockNo) == BlockState_BackUp));
        
        /* update counter of blocks marked for back up */
        --m_NumBlocksMarkedForBackup;
        
        /* set block state to free */
        m_BlockStateList.SetBlockStateAfterBackup (BlockNo, BlockState_Free);
        m_BlockStateList.SetBlockState (BlockNo, BlockState_Occupied);

        /* remove compression mode if there are no more blocks to compress */
        m_VolumeIsInCompressingMode = (m_NumBlocksMarkedForBackup != 0);
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }

    void Verify ();

public:
    
    FBM_ClusteredDataVolume (SAPDBMem_IRawAllocator &Allocator,
                    IOMan_BlockCount VolumeSize,
                    RTE_VolumeAccessMode VolMode)
        :
        FBM_IDataVolume(),
        m_LastBlock                            (VolumeSize-1),
        m_VolMode                              (VolMode),
        m_BlockStateList                       (Allocator),
        m_ClusterFreeList                      (Allocator, VolumeSize, 64),
        m_IsToDrop                             (SAPDB_FALSE),
        m_NumBlocksToAddUntilOptIsReached      (0),
        m_NumBlocksMarkedForBackup             (0),
        m_NumBlocksFreeAfterSVP                (0),
        m_NumBlocksUsed                        (0),
        m_VolumeIsInCompressingMode            (false) 
    {}

    void InitVolume();
    
    FBM_BlockState GetBlockState(IOMan_BlockNo BlockNo) const
    {
        return m_BlockStateList.GetBlockState(BlockNo); 
    }

    FBM_BlockState GetBlockStateAfterBackup(IOMan_BlockNo BlockNo) const
    {
        return  m_BlockStateList.GetBlockStateAfterBackup(BlockNo);  
    }

    void SetBlockStateAfterBackup(IOMan_BlockNo BlockNo, FBM_BlockState BlockState)
    { 
        m_BlockStateList.SetBlockStateAfterBackup (BlockNo, BlockState);
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }

    void SetBlockStateToBackUp (IOMan_BlockNo BlockNo)
    {
        SAPDBERR_ASSERT_STATE (!m_VolumeIsInCompressingMode);
        
        m_BlockStateList.SetBlockStateAfterBackup (BlockNo, BlockState_Occupied);
        m_BlockStateList.SetBlockState (BlockNo, BlockState_BackUp);
        ++ m_NumBlocksMarkedForBackup;
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }


    void SetBlockStateToFree (IOMan_BlockNo BlockNo)
    {
        m_BlockStateList.SetBlockState (BlockNo, BlockState_Free);

        m_ClusterFreeList.AddFreeBlock(BlockNo);

        --m_NumBlocksUsed;
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }
    
    void SetBlockStateToOccupied (IOMan_BlockNo BlockNo)
    {
        m_BlockStateList.SetBlockState (BlockNo, BlockState_Occupied);

        m_ClusterFreeList.RemoveFreeBlock(BlockNo);

        ++m_NumBlocksUsed;
        if (m_LastUsedBlock < BlockNo)
            m_LastUsedBlock = BlockNo;
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }
    
    void SetBlockStateToFreeAfterSVP (IOMan_BlockNo BlockNo)
    {
        /* set new block state */
        m_BlockStateList.SetBlockState (BlockNo, BlockState_FreeAfterSVP);
        
        /* update counter of blocks marked as free after savepoint */
        ++m_NumBlocksFreeAfterSVP;
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }

    
    void SetBlockStateMarkedForCompressionToFree (IOMan_BlockNo BlockNo)
    {
        pSetBlockStateMarkedForCompressionToFree (BlockNo);
        
        /* remove compression mode if there are no more blocks to compress */
        m_VolumeIsInCompressingMode = (m_NumBlocksMarkedForBackup != 0);
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }

    void SetBlockStateMarkedForCompressionToFreeAfterSVP (IOMan_BlockNo BlockNo);

    void SetBlockStateMarkedForCompressionToOccupied (IOMan_BlockNo BlockNo)
    {
        pSetBlockStateMarkedForCompressionToOccupied(BlockNo);
    }
    
    IOMan_BlockCount GetNumBlocks() const
    {
        return m_LastBlock + 1;
    }

    IOMan_BlockCount GetNumBlocksUsed() const
    {
        return m_NumBlocksUsed;
    }

    IOMan_BlockCount GetNumBlocksFree() const
    {
        if (!m_IsToDrop)
        {
            return m_LastBlock + 1 - m_NumBlocksUsed;
        }
        else  // simulate full volume, do not write!
        {
            return 0;
        }
    }

    void PrepareForDrop()
    {
        m_IsToDrop = SAPDB_TRUE;
    }

    void RestoreForDrop()
    {
        m_IsToDrop = SAPDB_FALSE;
    }

    SAPDB_Bool IsToDrop()
    {
        return m_IsToDrop;
    }

    IOMan_BlockCount GetNumBlocksToAddUntilOptIsReached() const
    {
        return m_NumBlocksToAddUntilOptIsReached;
    }

    void SetNumBlocksToAddUntilOptIsReached(IOMan_BlockCount value) 
    {
        m_NumBlocksToAddUntilOptIsReached = value;
    }


    IOMan_BlockNo GetActBlockNoForBackup() const
    {
        return m_ActBlockNoForBackup;
    }
        
    void SetActBlockNoForBackup(IOMan_BlockNo value) 
    {
        m_ActBlockNoForBackup = value;
    }
        
    IOMan_BlockCount GetNumBlocksMarkedForBackup() const
    {
        return m_NumBlocksMarkedForBackup;
    }

    IOMan_BlockCount GetNumBlocksFreeAfterSVP() const
    {
        return m_NumBlocksFreeAfterSVP;
    }

    RTE_VolumeAccessMode VolMode() const
    {
        return m_VolMode;
    }

    void RestoreAllBlockStatesMarkedForBackup (IOMan_BlockCount &NumBlocksRestoredToFreeAfterSVP);
    
    void RestoreBlockStateMarkedForBackup (
        IOMan_BlockNo  BlockNo,
        SAPDB_Bool &bRestoredStateIsFreeAfterSVP);
    
    void SetAllBlocksMarkedAsFreeAfterSVPToFree ();

    IOMan_BlockNo GetMultFreeBlocks (
        const IOMan_BlockCount  NumBlocksWanted,
        IOMan_BlockCount       &NumBlocksSupplied);

    IOMan_BlockNo GetFreeBlock ();

    void GetNextBlocksForBackUp (const IOMan_BlockCount   MaxNumBlocksWanted, 
        IOMan_BlockCount        &SuppliedNumBlocks, 
        IOMan_BlockNo        &BlockNo);

    void GetDeviceStatistics (IOMan_BlockCount &NumBlocks,
        IOMan_BlockCount &NumBlocksFree,
        IOMan_BlockCount &NumBlockFreeAfterSVP) const;

    /*--------------------------------------------------------------------------
    function:     Dump()
    description:  insert all state information of a data devspace into the 
                  kernel dump file
    arguments:    dump [in/out] kernel dump file
                  devNo [in] logical identifier for a data devspace
    return value: none
    --------------------------------------------------------------------------*/

    void Dump( 
        Kernel_Dump      &dump,
        const IOMan_DeviceNo devNo) const;

    /// dump all occupied blocks to knldiag
    void DumpUsedBlocks() const // dummy
    {}
};

#endif //FBM_CLUSTEREDDATAVOLUME_H

==> maxdb-source-7_6_00_27.tgz: V76_00_27/MaxDB_ORG/sys/src/SAPDB/FreeBlockManagement/FBM_ClusteredDataVolume.hpp <==
/*!

  @file           FBM_ClusteredDataVolume.hpp
  @author         Henrik
  @ingroup        FBM

\if EMIT_LICENCE
    ========== licence begin  GPL
    Copyright (c) 2002-2005 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

\endif
*/

#ifndef FBM_CLUSTEREDDATAVOLUME_H
#define FBM_CLUSTEREDDATAVOLUME_H

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "FreeBlockManagement/FBM_DataVolume.hpp"
#include "FreeBlockManagement/FBM_ClusteredFreeList.hpp"

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
  @class          FBM_ClusteredDataVolume
  @brief          representation of one data volume
 */
class FBM_ClusteredDataVolume : public FBM_DataVolume
{
private: 

    FBM_ClusteredFreeList m_ClusterFreeList;

    void Verify ();

public:
    
    /*!
        \brief   constructor for FBM_ClusteredDataVolume 
        \param   Allocator [in] allocator
        \param   VolumeSize [in] number of blocks in volume
        \param   VolMode    [in] volume mode
        \param   VolNo      [in] volume number (for debugging)
    */
    FBM_ClusteredDataVolume (SAPDBMem_IRawAllocator &Allocator,
                    IOMan_BlockCount        VolumeSize,
                    RTE_VolumeAccessMode    VolMode,
                    IOMan_VolumeNo          VolNo)
        :
        FBM_DataVolume(Allocator, VolumeSize, VolMode, VolNo),
        m_ClusterFreeList                      (Allocator, VolumeSize, 64)
    {}


    void InitVolume();

    void RestoreAllBlockStatesMarkedForBackup (IOMan_BlockCount &NumBlocksRestoredToFreeAfterSVP);

    void RestoreBlockStateMarkedForBackup (
        IOMan_BlockNo  BlockNo,
        SAPDB_Bool &bRestoredStateIsFreeAfterSVP);
    
    void SetAllBlocksMarkedAsFreeAfterSVPToFree ();

    void GetNextBlocksForBackUp (
        const IOMan_BlockCount  MaxNumBlocksWanted, 
        IOMan_BlockCount        &SuppliedNumBlocks, 
        IOMan_BlockNo           &BlockNo);

    IOMan_BlockNo GetMultFreeBlocks (
        const IOMan_BlockCount  NumBlocksWanted,
        IOMan_BlockCount       &NumBlocksSupplied);

    IOMan_BlockNo GetFreeBlock ();

    void SetBlockStateToFree (IOMan_BlockNo BlockNo)
    {
        m_BlockStateList.SetBlockState (BlockNo, BlockState_Free);

        m_ClusterFreeList.AddFreeBlock(BlockNo);

        --m_NumBlocksUsed;
        
#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }
    
    void SetBlockStateToOccupied (IOMan_BlockNo BlockNo)
    {
        m_BlockStateList.SetBlockState (BlockNo, BlockState_Occupied);
        ++m_NumBlocksUsed;

        if (m_LastUsedBlock < BlockNo)
        {
            m_LastUsedBlock = BlockNo;
        }
        m_ClusterFreeList.RemoveFreeBlock(BlockNo);

#       ifdef SAPDB_SLOW
        if( FBM_Check.ChecksLevel( 5 ))
            Verify ();
#       endif
    }

    // dummy procedure, used only by FBM_SimpleDataVolume
    bool getFreeCluster(
        const IOMan_BlockCount clusterSize,
        IOMan_BlockNo &clusterStart)
    {
        return false;
    };
    
    // dummy procedure, used only by FBM_SimpleDataVolume
    bool reserveClusterSpace(IOMan_ClusterCount numCluster)
    {
        return false;
    }

    // dummy procedure, used only by FBM_SimpleDataVolume
    IOMan_ClusterCount GetNumClustersFree() const
    {
        return 0;
    }


    void Dump( 
        Kernel_Dump          &dump,
        const IOMan_VolumeNo volNo) const;

    
};

#endif //FBM_CLUSTEREDDATAVOLUME_H
